@@ -1,3 +1,5 @@
+from . import onnx  # noqa: F401
+from .__about__ import __copyright__, __license__, __version__  # noqa: F401
 from .closure import Add, Constant, Lambda, Multiply, One, Unity, Zero  # noqa: F401
 from .container import (  # noqa: F401
     Broadcast,
@@ -13,7 +15,11 @@ from .conv import Conv1d, Conv2d, Conv3d  # noqa: F401
 from .convert import continual, forward_stepping  # noqa: F401
 from .delay import Delay  # noqa: F401
 from .linear import Linear  # noqa: F401
-from .module import CoModule, PaddingMode, TensorPlaceholder, call_mode  # noqa: F401
+from .module import CoModule, PaddingMode, call_mode  # noqa: F401
+from .multihead_attention import (  # noqa: F401
+    RetroactiveMultiheadAttention,
+    SingleOutputMultiheadAttention,
+)
 from .pooling import (  # noqa: F401
     AdaptiveAvgPool2d,
     AdaptiveAvgPool3d,
@@ -26,6 +32,14 @@ from .pooling import (  # noqa: F401
     MaxPool2d,
     MaxPool3d,
 )
+from .positional_encoding import RecyclingPositionalEncoding  # noqa: F401
 from .ptflops import _register_ptflops  # noqa: F401
+from .rnn import GRU, LSTM, RNN  # noqa: F401
 from .shape import Reshape  # noqa: F401
+from .transformer import (  # noqa: F401
+    RetroactiveTransformerEncoderLayer,
+    SingleOutputTransformerEncoderLayer,
+    TransformerEncoder,
+    TransformerEncoderLayerFactory,
+)
 from .utils import flat_state_dict, load_state_dict, state_dict  # noqa: F401
