@@ -1,4 +1,4 @@
-from typing import Callable, Tuple
+from typing import Callable, List, Optional, Tuple
 
 import torch
 import torch.nn.functional as F
@@ -15,18 +15,31 @@ from torch.nn.modules.conv import (
 )
 
 from .logging import getLogger
-from .module import CoModule, PaddingMode, TensorPlaceholder
+from .module import CoModule, PaddingMode
 
 logger = getLogger(__name__)
 
-State = Tuple[Tensor, int, int]
+# State = Tuple[Tensor, Optional[int], Optional[int]]
+State = Tuple[Tensor, Tensor, Tensor]
+
+# _forward_step_impl = None
+# from pathlib import Path
+# try:
+#     from torch.utils.cpp_extension import load as load_cpp
+
+#     _forward_step_impl = load_cpp(
+#         name="cpp_impl",
+#         sources=[str(Path(__file__).parent / "conv.cpp")],
+#         verbose=False,
+#     ).forward_step
+# except Exception as e:  # pragma: no cover
+#     logger.warning(
+#         "Unable to compile CoConv C++ implementation. Falling back to Python version."
+#     )
+#     logger.warning(e)
 
 
-__all__ = [
-    "Conv1d",
-    "Conv2d",
-    "Conv3d",
-]
+__all__ = ["Conv1d", "Conv2d", "Conv3d"]
 
 
 class _ConvCoNd(CoModule, _ConvNd):
@@ -52,14 +65,16 @@ class _ConvCoNd(CoModule, _ConvNd):
         assert issubclass(
             ConvClass, _ConvNd
         ), "The ConvClass should be a subclass of `_ConvNd`"
+
+        kernel_size = size_fn(kernel_size)
+        padding = size_fn(padding)
+        stride = size_fn(stride)
+
         self._ConvClass = ConvClass
         self._conv_func = conv_func
         self.input_shape_desciption = input_shape_desciption
         self._input_len = len(self.input_shape_desciption)
 
-        kernel_size = size_fn(kernel_size)
-        padding = size_fn(padding)
-        stride = size_fn(stride)
         if stride[0] > 1:
             logger.warning(
                 f"Temporal stride of {stride[0]} will result in skipped outputs every {stride[0]-1} / {stride[0]} steps"
@@ -101,60 +116,104 @@ class _ConvCoNd(CoModule, _ConvNd):
             self.kernel_size[0] - 1,
             *[0 for _ in self.padding[1:]],
         )
+        self._step_padding = (self.kernel_size[0] - 1, *self.padding[1:])
+        self._step_stride = (1, *self.stride[1:])
 
-        # init_state is called in `_forward_step`
+        self.register_buffer("state_buffer", torch.tensor([]), persistent=False)
+        self.register_buffer("state_index", torch.tensor(0), persistent=False)
+        self.register_buffer("stride_index", torch.tensor(0), persistent=False)
+
+    @property
+    def _stateless(self) -> bool:
+        return self.kernel_size[0] == 1 and self.padding[0] == 0 and self.stride[0] == 1
+
+    @property
+    def _state_shape(self) -> int:
+        if self._stateless:
+            return 0
+        else:
+            return 3
+
+    @property
+    def _dynamic_state_inds(self) -> List[bool]:
+        if self._stateless:
+            return []
+        else:
+            return [True, False, False]
 
     def init_state(
         self,
         first_output: Tensor,
     ) -> State:
         padding = self.make_padding(first_output)
-        state_buffer = padding.repeat(
-            self.kernel_size[0] - 1, *[1 for _ in self.input_shape_desciption]
+        repeat_shape = [self.kernel_size[0] - 1]
+        repeat_shape.extend((1,) * len(self.input_shape_desciption))
+        state_buffer = padding.repeat(repeat_shape)
+        state_index = torch.tensor(0)
+        stride_index = torch.tensor(
+            self.stride[0] - len(state_buffer) - 1 + self.padding[0]
         )
-        state_index = 0
-        stride_index = self.stride[0] - len(state_buffer) - 1 + self.padding[0]
-        if not hasattr(self, "state_buffer"):
-            self.register_buffer("state_buffer", state_buffer, persistent=False)
-        return state_buffer, state_index, stride_index
+        return (state_buffer, state_index, stride_index)
 
     def clean_state(self):
-        if hasattr(self, "state_buffer"):
-            del self.state_buffer
-        if hasattr(self, "state_index"):
-            del self.state_index
-        if hasattr(self, "stride_index"):
-            del self.stride_index
+        self.state_buffer = torch.tensor([])
+        self.state_index = torch.tensor(0)
+        self.stride_index = torch.tensor(0)
 
-    def get_state(self):
-        if (
-            hasattr(self, "state_buffer")
-            and self.state_buffer is not None
-            and hasattr(self, "state_index")
-            and self.state_index is not None
-            and hasattr(self, "stride_index")
-            and self.stride_index is not None
-        ):
+    def get_state(self) -> Optional[State]:
+        if len(self.state_buffer) > 0:
             return (self.state_buffer, self.state_index, self.stride_index)
+        return None
 
     def set_state(self, state: State):
         self.state_buffer, self.state_index, self.stride_index = state
 
-    def _forward_step(self, input: Tensor, prev_state: State) -> Tuple[Tensor, State]:
-        assert (
-            len(input.shape) == self._input_len - 1
-        ), f"A tensor of shape {(*self.input_shape_desciption[:2], *self.input_shape_desciption[3:])} should be passed as input but got {input.shape}"
+    @torch.jit.export
+    def _forward_step(
+        self, input: Tensor, prev_state: Optional[State]
+    ) -> Tuple[Optional[Tensor], Optional[State]]:
+        # assert (
+        #     len(input.shape) == self._input_len - 1
+        # ), f"A tensor of shape {(*self.input_shape_desciption[:2], *self.input_shape_desciption[3:])} should be passed as input but got {input.shape}"
 
+        # if (
+        #     _forward_step_impl is not None
+        #     and not self.training
+        #     and self.padding_mode == "zeros"
+        # ):
+        #     # Call C++ impl
+        #     output, next_state = _forward_step_impl(
+        #         input,
+        #         self.weight,
+        #         self.bias,
+        #         self.stride,
+        #         self.padding,
+        #         self._step_padding,
+        #         self.dilation,
+        #         self.groups,
+        #         *(prev_state or (None, None, None)),
+        #     )
+        #     if output is None:
+        #         output = None
+        #     return output, next_state
+        return self._forward_step_py(input, prev_state)
+
+    def _forward_step_py(
+        self, input: Tensor, prev_state: Optional[State]
+    ) -> Tuple[Optional[Tensor], Optional[State]]:
         # e.g. B, C -> B, C, 1
         x = input.unsqueeze(2).to(device=self.weight.device)
 
+        if self._stateless:
+            return self.forward(x).squeeze(2), prev_state
+
         if self.padding_mode == "zeros":
             x = self._conv_func(
                 input=x,
                 weight=self.weight,
                 bias=None,
-                stride=(1, *self.stride[1:]),
-                padding=(self.kernel_size[0] - 1, *self.padding[1:]),
+                stride=self._step_stride,
+                padding=self._step_padding,
                 dilation=self.dilation,
                 groups=self.groups,
             )
@@ -163,7 +222,7 @@ class _ConvCoNd(CoModule, _ConvNd):
                 input=F.pad(x, self._step_space_rprt, mode=self.padding_mode),
                 weight=self.weight,
                 bias=None,
-                stride=(1, *self.stride[1:]),
+                stride=self._step_stride,
                 padding=self._step_time_pad,
                 dilation=self.dilation,
                 groups=self.groups,
@@ -172,24 +231,39 @@ class _ConvCoNd(CoModule, _ConvNd):
         x_out, x_rest = x[:, :, 0], x[:, :, 1:]
 
         # Prepare previous state
-        buffer, index, stride_index = prev_state or self.init_state(x_rest)
+        if prev_state is None:
+            buffer, index, stride_index = self.init_state(x_rest)
+        else:
+            buffer, index, stride_index = prev_state
 
-        tot = len(buffer)
-        if stride_index == self.stride[0] - 1:
-            x_out = x_out.clone()
-            for i in range(tot):
-                x_out += buffer[(i + index) % tot, :, :, tot - i - 1]
+        assert index is not None
+        assert stride_index is not None
+
+        tot = self.kernel_size[0] - 1
+        output_is_valid = stride_index == self.stride[0] - 1
+
+        if output_is_valid:
+            x_out = x_out + (
+                torch.sum(
+                    buffer[
+                        torch.remainder(torch.arange(tot) + index, tot),
+                        :,
+                        :,
+                        torch.arange(tot - 1, -1, -1),
+                    ],
+                    dim=0,
+                )
+            )
 
             if self.bias is not None:
-                x_out += self.bias[
-                    (None, slice(None), *(None for _ in range(self._input_len - 3)))
-                ]
-        else:
-            x_out = TensorPlaceholder(x_out.shape)
+                bias = self.bias.unsqueeze(0)
+                for _ in range(self._input_len - 3):
+                    bias = bias.unsqueeze(-1)
+                x_out += bias
 
         # Update next state
         if self.kernel_size[0] > 1:
-            next_buffer = buffer.clone() if self.training else buffer.detach()
+            next_buffer = buffer.clone() if self.training else buffer
             next_buffer[index] = x_rest
             next_index = (index + 1) % tot
         else:
@@ -200,14 +274,17 @@ class _ConvCoNd(CoModule, _ConvNd):
         if next_stride_index > 0:
             next_stride_index = next_stride_index % self.stride[0]
 
-        return x_out, (next_buffer, next_index, next_stride_index)
+        if output_is_valid:
+            return x_out, (next_buffer, next_index, next_stride_index)
+        return None, (next_buffer, next_index, next_stride_index)
 
-    def forward_steps(self, input: Tensor, pad_end=False, update_state=True) -> Tensor:
-        assert (
-            len(input.shape) == self._input_len
-        ), f"A tensor of shape {self.input_shape_desciption} should be passed as input but got {input.shape}."
-
-        return CoModule.forward_steps(self, input, pad_end, update_state)
+    def forward_steps(
+        self, input: Tensor, pad_end: bool = False, update_state: bool = True
+    ) -> Optional[Tensor]:
+        # assert (
+        #     len(input.shape) == self._input_len
+        # ), f"A tensor of shape {self.input_shape_desciption} should be passed as input but got {input.shape}."
+        return self._forward_steps_impl(input, pad_end, update_state)
 
     def forward(self, input: Tensor) -> Tensor:
         """Performs a full forward computation exactly as the regular layer would.
@@ -222,7 +299,29 @@ class _ConvCoNd(CoModule, _ConvNd):
         assert (
             len(input.shape) == self._input_len
         ), f"A tensor of shape {self.input_shape_desciption} should be passed as input but got {input.shape}."
-        output = self._ConvClass._conv_forward(self, input, self.weight, self.bias)
+        # output = self._ConvClass._conv_forward(self, input, self.weight, self.bias)
+        if self.padding_mode == "zeros":
+            output = self._conv_func(
+                input=input,
+                weight=self.weight,
+                bias=self.bias,
+                stride=self.stride,
+                padding=self.padding,
+                dilation=self.dilation,
+                groups=self.groups,
+            )
+        else:
+            output = self._conv_func(
+                input=F.pad(
+                    input, self._reversed_padding_repeated_twice, mode=self.padding_mode
+                ),
+                weight=self.weight,
+                bias=self.bias,
+                stride=self.stride,
+                padding=(0,) * len(self.padding),
+                dilation=self.dilation,
+                groups=self.groups,
+            )
 
         return output
 
